[bits 16]
[org 0x7c00]

; colors
%define black   0
%define blue    1
%define green   2
%define cyan    3
%define red     4
%define purple  5
%define yellow  6
%define white   7

jmp word 0x0000:start

; Segments registers usage:
; fs - VGA buffer
; gs - global variables

%define globs     0x7e00

%define bbuffer   0       ; back buffer offset from gs, 80*50*2 bytes
%define bullets   80*50*2 ; 8 "bullets" in format: x[7]:y[7]:p[18]

; LCG
; s = a * s + c
; a = 0x41c64e6d
; c = 0x3039
; puts result in eax
rand:
  mov eax, [rseed]
  mov edx, 0x41c64e6d
  mul edx
  add eax, 0x3039
  mov [rseed], eax
  ret

; eax %= 24
; GCC says it works
; mod24:
;  push ecx
;  push edx
 
;  mov ecx, eax
;  mov edx, -1431655765
;  mul edx
;  mov eax, edx
;  shr eax, 4
;  lea edx, [eax+eax*2]
;  mov eax, ecx
;  sal edx, 3
;  sub eax, edx

;  pop edx
;  pop ecx
;  ret

; All calculation would be done on fixed-point integers in format 8:
start:
  ; Set stack segment register
  mov ax, 0x7bb
  mov sp, ax
  
  ; Set code segment register
;  mov ax, 0x7c0
;  mov cs, ax

  ; Setup fs to point into VGA buffer
  mov ax, 0xb800
  mov fs, ax
  
  ; Setup gs to point into some free memory
  ;mov ax, 0x7e00
  ;mov gs, ax

  ; Somehow cant do long jump to ml - need to ommit init code if it was alredy done
  mov al, [init_done]
  test al, al
  jnz ml

  ; Set 80x50 Text mode with 8x8 font
  mov ax, 0x1112
  xor bl, bl
  int 0x10

  ; Fill back buffer with some shiny color
  mov ecx, 80*50
  .fill:
    mov dword [globs+bbuffer+ecx*4-4], 0xffffffff
    loop .fill

  mov ecx, 8
  .rand_bullets:
    call rand
    and eax, ~(1 << 24)
    mov [globs+bullets+ecx*4-4], eax
    loop .rand_bullets

  mov byte [init_done], 1

ml:
  ; jmp swap_buffs
  ; jmp draw

  mov ecx, 80*50
  .fade_out:
    mov ax, [globs+bbuffer+ecx*2-2]
    sub ax, 40
    cmp ax, [globs+bbuffer+ecx*2-2]
    ja .fade_out_cont
    mov [globs+bbuffer+ecx*2-2], ax
  .fade_out_cont:
    loop .fade_out

  mov ecx, 8
  .animate:
    mov eax, [globs+bullets+ecx*4-4]
    mov ebx, eax
    and eax, 0x3ffff
    add eax, 100
    cmp eax, 0x3ffff
    ja .new_bullet
    and ebx, ~0x3ffff
    or eax, ebx
    jmp .update
    .new_bullet:
      call rand
      and eax, ~(1 << 24) & ~0x3ffff
    .update:
      mov [globs+bullets+ecx*4-4], eax
      loop .animate

.draw:
  mov ebx, 0
  .compute_bullets:
    mov ecx, 50
    mov edi, 80*50*2 - 2
    .cp_loop_y:
      ; pow(y-bullex.y, 2) -> esi
      mov eax, [globs+bullets+ebx*4]
      shr eax, 18
      and eax, 0x7f
      sub eax, ecx
      sub eax, ((64-50) / 2 + 1)
      imul eax
      push eax
      mov esi, eax

      push ecx
      mov ecx, 80
      .cp_loop_x:
        ; pow(y-bullet.y, 2) -> eax
        xor eax, eax
        mov eax, [globs+bullets+ebx*4]
        shr eax, 18+7
        and eax, 0x7f
        sub eax, ecx
        sub eax, ((128-80) / 2 + 1)
        imul eax

        ; pow(dist(pos, bullet), 4) - pos(dist(pos, bullt), 2)-> eax
        add eax, esi
        push eax
        mul eax
        pop edx
        shl edx, 8
        cmp edx, eax
        ja .cp_loop_x_nosub
        sub eax, edx
        
      .cp_loop_x_nosub:
        ; bullet.pow -> edx
        mov edx, [globs+bullets+ebx*4]
        and edx, 0x3ffff

        cmp eax, edx
        ja .cp_loop_x_cont

        sub edx, eax
        jmp .cp_loop_x_apply
        test edx, 0x20000
        jnz .cp_loop_x_cont

        test edx, 0x10000
        jz .cp_loop_x_apply
        mov dx, 0xffff

        .cp_loop_x_apply:
          cmp dx, [globs+bbuffer+edi]
          jb .cp_loop_x_cont
          mov [globs+bbuffer+edi], dx

        .cp_loop_x_cont:
          sub edi, 2
          loop .cp_loop_x

      pop ecx
      dec ecx
      jnz .cp_loop_y
      
    inc ebx
    cmp ebx, 8
    jb .compute_bullets

.swap_buffs:
  mov ecx, 80*50
  .draw_px:
    xor edx, edx
    mov dx, [globs+bbuffer+ecx*2-2]
    shr dx, (16 - 5)
    mov esi, edx
    and esi, 0x7
    mov al, [symbol+esi]
    mov esi, edx
    shr esi, 2
    and esi, 0x7
    mov ah, [bgrad+esi]
    mov [fs:ecx*2-2], ax
    loop .draw_px

  jmp word 0x0000:start

epilogue:
%if ($ -$$) > 510
  %fatal "Bootloader code too large"
%endif

symbol:
db 0, 0xb0, 0xb1, 0xb1, 0xdb, 0xb2, 0xb1, 0xb0

bgrad:
db black << 4 | 0x8 | black
db blue  << 4 | 0x8 | black
db blue  << 4 | 0x8 | blue
db cyan  << 4 | 0x8 | blue
db cyan  << 4 | 0x8 | green
db green << 4 | 0x8 | green
db green << 4 | 0x8 | yellow
db white << 4 | 0x8 | yellow

rseed:
  dd 0xfeedc0de

init_done:
  db 0

times 510 - ($ - $$) db 0xcc
db 0x55
db 0xaa
