; 12.06 23:59 - efekt graficzny jako bootloader (max 512 bajtw)
; Regs ax, bx, cx, dx
; PtrRegs ip, sp, bp
; IdxRegs si, di

%define loword(a) (a & 0xFFFF)
%define hiword(a) ((a >> 16) & 0xFFFF)
[bits 16]
[org 7c00h]
jmp word 0x0000:main
; - const_expr ----------------------------------------------------------------
%define STEP_TIME 1000 * 15
%define LH_TIME 1000 * 10
%define LH_COUNT 10
; - .bss ----------------------------------------------------------------------
%define font (7E00h) ; uint8_t[2 * 8]; 16 bytes
%define mersenne_mt (font + 16) ; uint32_t[624]; 2496 bytes
%define columns_cumul (mersenne_mt + 2496) ; uint16_t columns_cumul[80];
; - .data(==zero) --------------------------------------------------------------
mersenne_idx: dw 0
; - .rdata --------------------------------------------------------------------
sfg_square:
db 0b11111111
db 0b11000011
db 0b10111101
db 0b10100101
db 0b10100101
db 0b10111101
db 0b11000011
db 0b11111111
; - .code (functions) ---------------------------------------------------------
rand:
movzx eax, word [mersenne_idx]
cmp ax, 0
;jne _rand_main
; FIXME: generate numbers
; if (eax == 0)
;     for i from 0 to 623 {
;         int y := 32nd bit of(MT[i]) + last 31 bits of(MT[(i+1) mod 624])
;         MT[i] := MT[(i + 397) mod 624] xor (right shift by 1 bit(y))
;         if (y mod 2) == 1 { // y is odd
;             MT[i] := MT[i] xor (2567483615) // 0x9908b0df
;         }
;     }
; jmp $
;_rand_main:
mov ebx, [mersenne_mt + eax*4]
inc ax
xor dx, dx
mov cx, 624
div cx
mov [mersenne_idx], dx

mov eax, ebx
shr ebx, 11
xor eax, ebx

mov ebx, eax
shl ebx, 7
and ebx, 2636928640
xor eax, ebx

mov ebx, eax
shl ebx, 15
and ebx, 4022730752
xor eax, ebx

mov ebx, eax
shr ebx, 18
xor eax, ebx
ret

wait_step:
mov cx, hiword(STEP_TIME)
mov dx, loword(STEP_TIME)
jmp wait_sleep

wait_lh:
mov cx, hiword(LH_TIME)
mov dx, loword(LH_TIME)
;jmp wait_sleep

wait_sleep:
mov ah, 86h
int 15h ; Sleep (unknown time)
ret
; - .code (main) --------------------------------------------------------------
main:
; Init segments
xor ax, ax
mov ds, ax
mov es, ax
;mov fs, ax ; not-used
;mov gs, ax ; not-used

; Init stack
mov ss, ax
mov sp, 0x7C00

; Init RNG
; assume ax == 0
int 1Ah
mov ax, cx
shl eax, 16
mov ax, dx

mov di, mersenne_mt
stosd

mov ecx, 1
;mov di, mersenne_mt + 4

_mersenne_loop:
mov ebx, eax
shr ebx, 30
xor eax, ebx
mov ebx, 1812433253
mul ebx
add eax, ecx
stosd
inc ecx
cmp ecx, 624
jl _mersenne_loop

; Font generator
mov di, font ; set destination
; Write empty 8x8 char
xor ax, ax
mov cx, 8/2
rep stosw
; Write 8x8 square
mov si, sfg_square
mov cx, 8/2
rep movsw
; End of font generator

; Load 8x8 font
mov bp, font
mov ax, 1110h
mov cx, 2
xor dx, dx
xor bl, bl
mov bh, 8
int 10h
; end of font updateing

; Init columns_cumul
mov ax, 50
mov di, columns_cumul
mov cx, 80
init_cc:
stosw
add ax, 50
loop init_cc

; Set segment es to text mem
mov ax, 0xB800 ; 0xb8000
mov es, ax

; Clear screen
xor di, di
xor ax, ax
mov cx, 80 * 50 * 2 / 2
rep stosw

_process_board:
; --- Test if last-line filled ---
mov di, 2 * 49 * 80 ; last line
mov cx, 81	; Check all chars
xor ax, ax
repne scasw ; Find first null
test cx, cx
jnz _process_gravity
; --- Fall down code start ----------------------------------------------------
; --- Highlight lower-line green ---
mov di, 2 * 49 * 80 + 1 ; left attr
mov si, 2 * 50 * 80 + 1 - 2; right attr
mov cx, 40	; Check all chars
mov al, 0b00000010
_line_highlight:
stosb
inc di
mov byte [es:si], al
sub si, 2
pusha
call wait_lh
popa
loop _line_highlight
; --- End of highlighting ---
; --- Blink last line several times ---
mov cx, LH_COUNT
_blink_line_loop:
push cx
mov di, 2 * 49 * 80 + 1 ; left attr
mov cx, 80 ; Check all chars
mov ax, 0x0000
rep stosw
call wait_lh
mov di, 2 * 49 * 80 + 1 ; left attr
mov cx, 80 ; Check all chars
mov ax, 0x0102
rep stosw
call wait_lh
pop cx
loop _blink_line_loop
; --- End of blinking ---
; --- Main falling code ---
mov si, (49 * 80 - 1) * 2 ; di = board.back() - board.width();
mov di, (50 * 80 - 1) * 2 ; di = board.back();
mov cx, 49 * 80 + 1
push ds
mov ax, es
mov ds, ax
std
rep movsw
cld
pop ds
; --- End of falling code ---
; --- Clear 1st line ---
xor di, di
mov cx, 80
xor ax, ax
rep stosw
; --- Line cleared ---
; --- Incrase columns_cumul ---
mov si, columns_cumul
mov cx, 80
inc_cc:
inc ax
add [ds:si], ax
add si, 2
loop inc_cc
; --- End of incrasing ---
; --- Fall down code end ------------------------------------------------------
; --- Gravity tick ------------------------------------------------------------
_process_gravity:
std
mov di, (50 * 80 - 1) * 2 ; di = board.back()
mov cx, (50 - 1) * 80 + 1 ; cx = (board.size() - board.line_size()) + 1
;xor ax, ax ; assume == 0
xor dx, dx

__pg_loop:
repne scasw
test cx, cx
jz __pg_end ; End of board reached, goto wait
cmp word [es:di-160+2], 0 
jz __pg_loop
xchg dx, [es:di-160+2]
;mov word [es:di-160+2], 0
xchg [es:di+2], dx
jmp __pg_loop

__pg_end:
cld
; --- Gravity tick end --------------------------------------------------------
; --- Inserting brick at random position on the top line ----------------------
call rand
mov bh, ah ; Save rand for char color
xor dx, dx
mov cx, [columns_cumul+2*79]
dec cx
div cx
inc dx
mov ax, dx
mov si, columns_cumul - 2
_column_cumul_loop:
add si, 2
cmp ax, [ds:si]
jae _column_cumul_loop ; if (ax >= *ds:++si)
mov di, si
sub di, columns_cumul

mov cx, di
shr cx, 1
neg cx
add cx, 80

dec_cc:
dec word [ds:si]
add si, 2
loop dec_cc

; --- Write character ---
and bh, 0b00001111
setz bl
add bh, bl
mov bl, 1
mov [es:di], bx

call wait_step
jmp _process_board

times 510 - ($ - $$) db 0
db 0x55
db 0xAA